# Copyright 2023 The MIT Probabilistic Computing Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass

from jax import api_util
from jax import core as jax_core
from jax import tree_util as jtu
from jax._src import dtypes
from jax.extend import linear_util as lu
from jax.interpreters import partial_eval as pe
from jax.util import safe_map

from genjax._src.core.pytree.pytree import Pytree
from genjax._src.core.typing import Any


@dataclass
class Concrete(Pytree):
    v: Any

    def flatten(self):
        return (), (self.v,)

    def __hash__(self):
        return hash(self.v)


def concrete(v):
    return Concrete(v)


def static_check_concrete(v):
    return isinstance(v, Concrete)


def static_unwrap_concrete(c):
    if isinstance(c, Concrete):
        return c.v
    else:
        return c


def get_shaped_aval(x):
    if hasattr(x, "dtype") and hasattr(x, "shape"):
        return jax_core.ShapedArray(x.shape, dtypes.canonicalize_dtype(x.dtype))
    return jax_core.raise_to_shaped(jax_core.get_aval(x))


@lu.cache
def cached_stage_dynamic(flat_fun, in_avals):
    jaxpr, _, consts = pe.trace_to_jaxpr_dynamic(flat_fun, in_avals)
    typed_jaxpr = jax_core.ClosedJaxpr(jaxpr, consts)
    return typed_jaxpr


def stage(f):
    """Returns a function that stages a function to a ClosedJaxpr."""

    def wrapped(*args, **kwargs):
        fun = lu.wrap_init(f, kwargs)
        flat_args, in_tree = jtu.tree_flatten(args)
        flat_fun, out_tree = api_util.flatten_fun_nokwargs(fun, in_tree)
        flat_avals = safe_map(get_shaped_aval, flat_args)
        typed_jaxpr = cached_stage_dynamic(flat_fun, tuple(flat_avals))
        return typed_jaxpr, (flat_args, in_tree, out_tree)

    return wrapped


def trees(f):
    """Returns a function that determines input and output pytrees from inputs,
    and also returns the flattened input arguments."""

    def wrapped(*args, **kwargs):
        return stage(f)(*args, **kwargs)[1]

    return wrapped
